--[[
Copyright (C) GtX (Andy), 2021

Author: GtX | Andy
Date: 19.01.2021
Revision: FS22-01

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Important:
Not to be added to any mods / maps or modified from its current release form.
No modifications may be made to this script, including conversion to other game versions without written permission from GtX | Andy
Copying or removing any part of this code for external use without written permission from GtX | Andy is prohibited.

Darf nicht zu Mods / Maps hinzugefügt oder von der aktuellen Release-Form geändert werden.
Ohne schriftliche Genehmigung von GtX | Andy dürfen keine Änderungen an diesem Skript vorgenommen werden, einschließlich der Konvertierung in andere Spielversionen
Das Kopieren oder Entfernen irgendeines Teils dieses Codes zur externen Verwendung ohne schriftliche Genehmigung von GtX | Andy ist verboten.
]]


SleeperCabManager = {}

SleeperCabManager.xmlSchema = nil
SleeperCabManager.numInserted = 0

local SleeperCabManager_mt = Class(SleeperCabManager)

local validationFail
local developmentMode = false

local buildId = 1
local versionString = "0.0.0.0"

local modName = g_currentModName
local modDirectory = g_currentModDirectory

g_xmlManager:addCreateSchemaFunction(function ()
    SleeperCabManager.xmlSchema = XMLSchema.new("globalSleeperCab")
end)

g_xmlManager:addInitSchemaFunction(function ()
    local schema = SleeperCabManager.xmlSchema

    schema:register(XMLValueType.STRING, "vehicleSleeperCab.storeItem(?)#version", "Configured Version", "1.0.0.0")
    schema:register(XMLValueType.STRING, "vehicleSleeperCab.storeItem(?)#xmlFilename", "XML filename")

    schema:register(XMLValueType.BOOL, "vehicleSleeperCab.storeItem(?)#deactivateEngineOnSleep", "Should deactivate vehicle engine when sleeping starts", true)
    schema:register(XMLValueType.BOOL, "vehicleSleeperCab.storeItem(?)#deactivateLightsOnSleep", "Should deactivate vehicle lights when sleeping starts", true)

    schema:register(XMLValueType.STRING, "vehicleSleeperCab.storeItem(?).requiredConfiguration#name", "Configuration name required indexes")
    schema:register(XMLValueType.INT, "vehicleSleeperCab.storeItem(?).requiredConfiguration.configuration(?)#index", "Configuration index to allow sleeping")
    schema:register(XMLValueType.VECTOR_TRANS, "vehicleSleeperCab.storeItem(?).requiredConfiguration.configuration(?)#cameraTranslation", "Camera translation when index active (Optional)")
    schema:register(XMLValueType.VECTOR_ROT, "vehicleSleeperCab.storeItem(?).requiredConfiguration.configuration(?)#cameraRotation", "Camera rotation when index active (Optional)")

    schema:register(XMLValueType.VECTOR_TRANS, "vehicleSleeperCab.storeItem(?).camera#translation", "Camera translation")
    schema:register(XMLValueType.VECTOR_ROT, "vehicleSleeperCab.storeItem(?).camera#rotation", "Camera rotation")
    schema:register(XMLValueType.ANGLE, "vehicleSleeperCab.storeItem(?).camera#fovy", "Camera field of view", 60)
    schema:register(XMLValueType.FLOAT, "vehicleSleeperCab.storeItem(?).camera#nearClip", "Camera near clip", 0.1)
    schema:register(XMLValueType.FLOAT, "vehicleSleeperCab.storeItem(?).camera#farClip", "Camera far clip", 5000)

    schema:setXMLSpecializationType()
end)

function SleeperCabManager.new()
    self = setmetatable({}, SleeperCabManager_mt)

    self.modName = modName
    self.modDirectory = modDirectory

    self.buildId = buildId
    self.versionString = versionString

    self.globalStoreItems = {}
    self.vehicles = {}

    self.xmlFilename = modDirectory .. "resources/globalVehicles.xml"

    if buildId == 0 then
        self:enableDevelopmentMode()
    end

    return self
end

function SleeperCabManager:addVehicle(vehicle)
    if self.vehicles ~= nil then
        self.vehicles[vehicle] = vehicle
    end
end

function SleeperCabManager:removeVehicle(vehicle)
    if self.vehicles ~= nil then
        self.vehicles[vehicle] = nil
    end
end

function SleeperCabManager:getIsGlobalVehicle(configFileName)
    if configFileName == nil then
        return false
    end

    return self.globalStoreItems[configFileName:lower()] == true
end

function SleeperCabManager:enableDevelopmentMode()
    local addCommand = not developmentMode

    developmentMode = g_server ~= nil

    if addCommand and developmentMode then
        -- No 'removeConsoleCommand' as this is DEVELOPMENT only
        addConsoleCommand("gtxSleeperCabReloadGlobalVehicles", "Reloads the global vehicle XML file [reloadVehicles]", "consoleCommandReloadGlobalVehicles", self)
    end

    return developmentMode
end

function SleeperCabManager:consoleCommandReloadGlobalVehicles(reloadVehicles)
    if SleeperCabManager.numInserted > 0 then
        local numGlobalVehicles = SleeperCabManager.loadGlobalVehiclesFromXML()

        if numGlobalVehicles > 0 then
            if reloadVehicles == "true" then
                g_currentMission:consoleCommandReloadVehicle("false", "100")
            end
        end

        return "Global XML reloaded, Added " .. tostring(numGlobalVehicles) .. " new vehicles"
    end

    return "Failed to reload global XML"
end

function SleeperCabManager.loadGlobalVehiclesFromXML()
    local numGlobalVehicles = 0
    local manager = g_sleeperCabManager

    if manager ~= nil then
        local xmlFile = XMLFile.loadIfExists("SleeperCabXML", manager.xmlFilename, SleeperCabManager.xmlSchema)

        if xmlFile ~= nil then
            xmlFile:iterate("vehicleSleeperCab.storeItem", function (_, key)
                local xmlFilename = xmlFile:getValue(key .. "#xmlFilename")

                if xmlFilename ~= nil then
                    local storeitem = g_storeManager:getItemByXMLFilename(NetworkUtil.convertFromNetworkFilename(xmlFilename))

                    if storeitem ~= nil and manager.globalStoreItems[storeitem.xmlFilenameLower] == nil then
                        manager.globalStoreItems[storeitem.xmlFilenameLower] = true

                        numGlobalVehicles = numGlobalVehicles + 1
                    end
                end
            end)

            xmlFile:delete()
        end
    end

    return numGlobalVehicles
end

local function getIsSleeperVehicle(vehicle)
    if vehicle == nil or vehicle.getIsSleeper == nil then
        return false
    end

    return vehicle:getIsSleeper()
end

local function typeManagerValidateTypes(self)
    if self.typeName == "vehicle" then
        local specializationName = string.format("%s.sleeperCab", modName)
        local specializationObject = g_specializationManager:getSpecializationObjectByName(specializationName)

        if specializationObject ~= nil then
            if g_specializationManager.sortedSpecializations ~= nil then
                for _, spec in ipairs (g_specializationManager.sortedSpecializations) do
                    if spec.className ~= nil and string.endsWith(spec.className, ".ChangeEnterText") then
                        local specialization = g_specializationManager:getSpecializationObjectByName(spec.name)

                        if specialization ~= nil and specialization.onLoad ~= nil and rawget(specialization, "updateActionEvents") ~= nil then
                            specialization.onLoad = function()
                                -- print("[ChangeEnterText] Disabled due to conflicts, and no superFunc return. Also breaks bale triggers!!")
                            end
                        end
                    end
                end
            end

            for typeName, typeEntry in pairs (self:getTypes()) do
                if specializationObject.prerequisitesPresent(typeEntry.specializations) then
                    self:addSpecialization(typeName, specializationName)

                    SleeperCabManager.numInserted = SleeperCabManager.numInserted + 1
                end
            end
        end
    end
end

local function playerNew(self, superFunc, ...)
    local instance = superFunc(self, ...)

    if instance.inputInformation ~= nil then
        local onInputSleep = function(playerObject, actionName, inputValue, callbackState, isAnalog, isMouse, deviceCategory)
            local vehicle = g_currentMission.interactiveVehicleInRange

            if vehicle ~= nil and vehicle.openSleepDialog ~= nil then
                vehicle:openSleepDialog(playerObject)
            end
        end

        if developmentMode and g_server ~= nil then
            local cameraNode = nil

            onInputSleep = function(playerObject, actionName, inputValue, callbackState, isAnalog, isMouse, deviceCategory)
                local sleeperCabInput = playerObject.inputInformation.registrationList[InputAction.VEHICLE_SLEEP]

                if cameraNode == nil then
                    local vehicle = g_currentMission.interactiveVehicleInRange

                    if vehicle ~= nil and vehicle.getSleepCamera ~= nil then
                        local sleepCamera = vehicle:getSleepCamera()

                        if sleepCamera ~= nil then
                            cameraNode = getCamera()
                            setCamera(sleepCamera)

                            if sleeperCabInput ~= nil then
                                g_inputBinding:setActionEventText(sleeperCabInput.eventId, "Disable Sleep Camera Debug")
                            end
                        else
                            vehicle:openSleepDialog(playerObject)
                        end
                    end
                else
                    if playerObject.isEntered then
                        setCamera(cameraNode)
                    end

                    cameraNode = nil

                    if sleeperCabInput ~= nil then
                        g_inputBinding:setActionEventText(sleeperCabInput.eventId, "Enable Sleep Camera Debug")
                    end
                end

                g_currentMission.hud:setIsVisible(cameraNode == nil)
            end

            print("  Info: [VehicleSleeperCab] Development mode 'onInputSleep' function activated for player!")
        end

        instance.inputInformation.registrationList[InputAction.VEHICLE_SLEEP] = {
            text = g_i18n:getText("ui_inGameSleep"),
            triggerAlways = false,
            triggerDown = true,
            eventId = "",
            textVisibility = false,
            triggerUp = false,
            callback = onInputSleep,
            activeType = Player.INPUT_ACTIVE_TYPE.STARTS_DISABLED
        }
    end

    return instance
end

local function playerUpdateActionEvents(self)
    local sleeperCabInput = self.inputInformation.registrationList[InputAction.VEHICLE_SLEEP]

    if sleeperCabInput ~= nil then
        local vehicleSleepInputState = self.canEnterVehicle and getIsSleeperVehicle(g_currentMission.interactiveVehicleInRange)

        g_inputBinding:setActionEventTextVisibility(sleeperCabInput.eventId, vehicleSleepInputState)
        g_inputBinding:setActionEventActive(sleeperCabInput.eventId, vehicleSleepInputState)
    end
end

local function farmGetSleepCamera(self, superFunc)
    if not self.isSpectator and g_sleeperCabManager and g_sleeperCabManager.vehicles ~= nil then
        if g_currentMission.controlledVehicle == nil and g_currentMission.enteredPassengerVehicle == nil then
            local sleeperVehicle = nil
            local maxVehicleDistance = 20
            local closestSleeperVehicle = math.huge

            local accessHandler = g_currentMission.accessHandler
            local playerNode = g_currentMission.player ~= nil and g_currentMission.player.rootNode or nil

            for _, vehicle in pairs(g_sleeperCabManager.vehicles) do
                if not vehicle.isDeleted and not vehicle.isDeleting and not vehicle.isControlled then
                    if playerNode ~= nil and not vehicle.isBroken and getIsSleeperVehicle(vehicle) and accessHandler:canFarmAccessOtherId(self.farmId, vehicle:getOwnerFarmId()) then
                        local distanceToVehicle = calcDistanceFrom(vehicle.rootNode, playerNode)

                        if distanceToVehicle <= maxVehicleDistance and distanceToVehicle < closestSleeperVehicle then
                            closestSleeperVehicle = distanceToVehicle
                            sleeperVehicle = vehicle
                        end
                    end
                else
                    g_sleeperCabManager:removeVehicle(vehicle)
                end
            end

            -- If the player is within 20m of a sleeper cab vehicle use it
            if getIsSleeperVehicle(sleeperVehicle) then
                return sleeperVehicle:getSleepCamera()
            end
        end
    end

    return superFunc(self)
end

local function sleepManagerLoadMapData(self)
    if SleeperCabManager.numInserted > 0 then
        g_asyncTaskManager:addTask(function ()
            local numGlobalVehicles = SleeperCabManager.loadGlobalVehiclesFromXML()

            if numGlobalVehicles > 0 then
                Logging.info("[VehicleSleeperCab] - Version %s has been loaded into %d store items successfully", versionString, numGlobalVehicles)
            end
        end)
    end
end

local function sleepManagerStartSleep(self, superFunc, ...)
    local player = g_currentMission.player
    local camera = self:getCamera()
    local x, y, z, rx, ry, rz = nil

    if camera ~= nil and camera ~= self.fallbackCamera then
        local useCamera = getUserAttribute(camera, "useCamera") -- Allow other mods including placeables to use this feature if they have a nice camera position they wish to use

        if useCamera ~= nil then
            if useCamera == true then
                x, y, z = getWorldTranslation(camera)
                rx, ry, rz = getWorldRotation(camera)
            end
        elseif player ~= nil and g_modIsLoaded["FS22_LogCabinFarmhouse"] then -- Set my log cabin directly if in use to remove need for an update :-)
            local farm = g_farmManager:getFarmById(player.farmId)
            local farmhouse = farm ~= nil and farm:getFarmhouse() or nil

            if farmhouse ~= nil and farmhouse:getSleepCamera() == camera and farmhouse.customEnvironment == "FS22_LogCabinFarmhouse" then
                setUserAttribute(camera, "useCamera", "Boolean", true)

                x, y, z = getWorldTranslation(camera)
                rx, ry, rz = getWorldRotation(camera)
            end
        end
    end

    superFunc(self, ...)

    if x ~= nil and camera ~= nil and camera == getCamera() then
        -- Correct the camera and hide the overlay, not sure why this is not an option
        setWorldTranslation(camera, x, y, z)
        setWorldRotation(camera, rx, ry, rz)
    end

    if g_sleeperCabManager ~= nil and g_sleeperCabManager.vehicles ~= nil then
        for _, vehicle in pairs(g_sleeperCabManager.vehicles) do
            if not vehicle.isBroken and vehicle.onStartSleep ~= nil then
                vehicle:onStartSleep()
            end
        end
    end

    if player ~= nil and player.isEntered and player.isClient then
        player:setLightIsActive(false)
    end
end

local function sleepManagerStopSleep(self)
    local player = g_currentMission.player

    if player == nil or g_currentMission.isPlayerFrozen or g_gui:getIsGuiVisible()then
        return
    end

    if player.isEntered and player.isClient then
        player:updateActionEvents()
    end
end

local function init()
    if g_globalMods ~= nil and g_globalMods.sleeperCabManager ~= nil then
        Logging.error("Failed to load mod '%s', 'SleeperCabManager.lua' was already loaded by '%s'!", modName, g_globalMods.vehicleSpeedSync.modName or "N/A")

        return
    end

    local isValidMod = false
    local mod = g_modManager:getModByName(modName)

    versionString = mod.version or versionString

    if mod.modName == "FS22_VehicleSleeperCab" or mod.modName == "FS22_VehicleSleeperCab_update" then
        isValidMod = mod.author and #mod.author == 3
    end

    if isValidMod then
        g_sleeperCabManager = SleeperCabManager.new()

        if g_globalMods ~= nil then
            g_globalMods.sleeperCabManager = g_sleeperCabManager

            if g_globalMods.easyDevControls ~= nil and g_globalMods.easyDevControls.godMode ~= nil then
                g_globalMods.easyDevControls.godMode:enableModDevelopment("FS22_VehicleSleeperCab", g_sleeperCabManager)
            end
        end

        g_specializationManager:addSpecialization("sleeperCab", "SleeperCab", modDirectory .. "scripts/specializations/SleeperCab.lua", nil)

        TypeManager.validateTypes = Utils.prependedFunction(TypeManager.validateTypes, typeManagerValidateTypes)

        Player.new = Utils.overwrittenFunction(Player.new, playerNew)
        Player.updateActionEvents = Utils.appendedFunction(Player.updateActionEvents, playerUpdateActionEvents)

        Farm.getSleepCamera = Utils.overwrittenFunction(Farm.getSleepCamera, farmGetSleepCamera)

        SleepManager.loadMapData = Utils.appendedFunction(SleepManager.loadMapData, sleepManagerLoadMapData)
        SleepManager.startSleep = Utils.overwrittenFunction(SleepManager.startSleep, sleepManagerStartSleep)
        SleepManager.stopSleep = Utils.prependedFunction(SleepManager.stopSleep, sleepManagerStopSleep)
    else
        validationFail = {
            startUpdateTime = 2000,

            update = function(self, dt)
                self.startUpdateTime = self.startUpdateTime - dt

                if self.startUpdateTime < 0 then
                    local text = string.format(g_i18n:getText("vehicleSleeperCab_loadError", modName), modName, mod.author or "Unknown")

                    if g_dedicatedServerInfo == nil then
                        if not g_gui:getIsGuiVisible() then
                            g_gui:showYesNoDialog({
                                title = string.format("%s - Version %s", mod.title, versionString),
                                text = text,
                                dialogType = DialogElement.TYPE_LOADING,
                                callback = self.openModHubLink,
                                yesText = g_i18n:getText("button_ok"),
                                noText = g_i18n:getText("button_modHubDownload")
                            })
                        end
                    else
                        print("\n" .. text .. "\n    - https://farming-simulator.com/mods.php?lang=en&country=be&title=fs2022&filter=org&org_id=129652&page=0" .. "\n")
                        self.openModHubLink(true)
                    end
                end
            end,

            openModHubLink = function(ignore)
                if ignore == false then
                    local language = g_languageShort
                    local link = "mods.php?lang=en&country=be&title=fs2022&filter=org&org_id=129652&page=0"
                    if language == "de" or language == "fr" then
                        link = "mods.php?lang=" .. language .. "&country=be&title=fs2022&filter=org&org_id=129652&page=0"
                    end

                    openWebFile(link, "")
                end

                removeModEventListener(validationFail)
                validationFail = nil
            end
        }

        addModEventListener(validationFail)
    end
end

init()
